from typing import Callable

from lenses import lens

from amino import (List, _, __, Just, Maybe, Map, Right, Empty, Either,
                   Boolean, I)
from amino.lazy import lazy
from amino.task import Task
from amino.anon import L

from ribosome.machine import may_handle, handle, Quit, Nop
from ribosome.record import field, list_field, Record
from ribosome.machine.base import UnitTask, DataTask

from myo.state import MyoComponent, MyoTransitions
from myo.plugins.tmux.message import (TmuxOpen, TmuxRunCommand,
                                      TmuxCreatePane, TmuxCreateLayout,
                                      TmuxCreateSession, TmuxSpawnSession,
                                      TmuxFindVim, TmuxInfo, TmuxLoadDefaults,
                                      TmuxClosePane, TmuxRunShell,
                                      TmuxRunLineInShell, StartWatcher,
                                      WatchCommand, QuitWatcher, SetCommandLog,
                                      TmuxPack, TmuxPostOpen, TmuxFixFocus,
                                      TmuxMinimize, TmuxRestore, TmuxToggle,
                                      TmuxFocus, TmuxOpenOrToggle)
from myo.plugins.core.main import StageI
from myo.ui.tmux.pane import Pane, VimPane
from myo.ui.tmux.layout import LayoutDirections, Layout, VimLayout
from myo.ui.tmux.session import Session
from myo.ui.tmux.server import Server, NativeServer
from myo.util import parse_int, view_params, pane_params, pane_bool_params
from myo.ui.tmux.window import VimWindow, Window
from myo.ui.tmux.facade import LayoutFacade, PaneFacade, ViewFacade
from myo.ui.tmux.view import View
from myo.plugins.core.message import AddDispatcher
from myo.plugins.tmux.dispatch import TmuxDispatcher
from myo.ui.tmux.util import format_state, Ident
from myo.ui.tmux.view_path import LayoutPathMod, PanePathMod, ViewPathMod
from myo.plugins.command.message import SetShellTarget
from myo.command import Command, Shell
from myo.plugins.tmux.watcher import Watcher, Terminated

_is_vim_window = lambda a: isinstance(a, VimWindow)
_is_vim_layout = lambda a: isinstance(a, VimLayout)
_is_vim_pane = lambda a: isinstance(a, VimPane)


def _ident_ppm(ident: Ident):
    return PanePathMod(pred=__.has_ident(ident))


def _ident_lpm(ident: Ident):
    return LayoutPathMod(pred=__.has_ident(ident))


def _ident_vpm(ident: Ident):
    return ViewPathMod(pred=__.has_ident(ident))


class TmuxState(Record):
    server = field(Server)
    sessions = list_field()
    windows = list_field()
    instance_id = field(str, initial='', factory=L(List.random_string)(5))

    @property
    def vim_window(self):
        return self.windows.find(_is_vim_window)

    @property
    def vim_window_lens(self):
        return (self.windows.find_lens_pred(_is_vim_window) /
                lens(self).windows.add_lens)

    @property
    def vim_layout(self):
        return self.vim_window // __.root.layouts.find(_is_vim_layout)

    @property
    def vim_layout_lens(self):
        return (
            self.vim_layout //
            __.layouts.find_lens_pred(_is_vim_layout) /
            lens(self).windows.add_lens
        )

    @property
    def vim_pane(self):
        return self.vim_layout / _.panes / _.head

    @property
    def all_panes(self):
        return self.windows // _.root.all_panes

    def pane(self, ident):
        return self.all_panes.find(__.has_ident(ident))

    @property
    def possibly_open_panes(self):
        return (
            self.all_panes
            .filter(_.id.is_just)
            .filter(lambda a: not _is_vim_pane(a))
        )

    def find_view(self, view: View):
        return self.windows.find_map(__.root.find_view(view))


class TmuxTransitions(MyoTransitions):

    def _with_root(self, data, state, root):
        l = state.vim_window_lens / _.root
        new_state = l / __.set(root) | state
        return self._with_sub(data, new_state)

    def _with_vim_window(self, data, state, win):
        new_state = state.vim_window_lens / __.set(win) | state
        return self._with_sub(data, new_state)

    def with_sub_and_msgs(self, state, msgs):
        return (self.with_sub(state),) + msgs

    @property
    def server(self):
        return self.state.server

    @property
    def layouts(self):
        return self.machine.layouts

    @property
    def views(self):
        return self.machine.views

    @property
    def panes(self):
        return self.machine.panes

    @may_handle(StageI)
    def stage_1(self):
        ''' Initialize the state. If it doesn't exist, Env will create
        it using the constructor function supplied in *Plugin.state*
        '''
        default = self.pflags.tmux_use_defaults.maybe(TmuxLoadDefaults())
        msgs = List(AddDispatcher(), self.with_sub(self.state), TmuxFindVim())
        return msgs + default.to_list + List(StartWatcher())

    @may_handle(AddDispatcher)
    def add_dispatcher(self):
        return self.data + TmuxDispatcher()

    @may_handle(TmuxFindVim)
    def find_vim(self):
        vim_pid = self._vim_pid
        vim_pane = vim_pid // self._find_vim_pane
        id = (
            (self.vim.vars.p('tmux_force_vim_pane_id') // parse_int)
            .or_else(vim_pane // _.id_i)
        )
        wid = (
            self.vim.vars.p('tmux_force_vim_win_id')
            .or_else(vim_pane // _.window_id_i)
            .to_either('no vim win id')
        )
        sid = (
            self.vim.vars.p('tmux_force_vim_session_id')
            .or_else(vim_pane // _.session_id_i)
            .to_either('no vim session id')
        )
        vim_w = self.vim.vars.p('tmux_vim_width').or_else(Just(88))
        pane = VimPane(id=id.to_maybe, name='vim', pid=vim_pid.to_maybe,
                       window_id=wid, session_id=sid)
        vim_layout = VimLayout(name='vim',
                               direction=LayoutDirections.vertical,
                               panes=List(pane), min_size=vim_w,
                               weight=Just(0.0001))
        root = Layout(name='root',
                      direction=LayoutDirections.horizontal,
                      layouts=List(vim_layout))
        win = VimWindow(id=wid.to_maybe, root=root)
        new_state = self.state.append1.windows(win)
        return self.with_sub(new_state)

    @may_handle(TmuxLoadDefaults)
    def load_defaults(self):
        main = TmuxCreateLayout('main',
                                options=Map(parent='root', weight=1))
        make = TmuxCreatePane('make', options=Map(parent='main', pin=True,
                                                  position=-1))
        return main, make

    @may_handle(StartWatcher)
    def start_watcher(self):
        return UnitTask(Task(self.machine.watcher.start))

    @handle(TmuxCreateSession)
    def create_session(self):
        return (
            self.msg.options.get('name') /
            (lambda n: Session(id=n)) /
            List /
            self.state.append.sessions /
            self.with_sub
        )

    @may_handle(TmuxSpawnSession)
    def spawn_session(self):
        s = self.state.session(self.msg.id)
        self.server.new_session(session_name=s.x)

    @handle(TmuxCreateLayout)
    def create_layout(self):
        opts = self.msg.options
        dir_s = opts.get('direction') | 'vertical'
        direction = LayoutDirections.parse(dir_s)
        layout = Layout(name=self.msg.name, direction=direction,
                        **view_params(opts))
        return self._add_to_layout(opts.get('parent'), _.layouts, layout)

    @handle(TmuxCreatePane)
    def create_pane(self):
        opts = self.msg.options
        pane = Pane(name=self.msg.name, **pane_params(opts),
                    **pane_bool_params(opts))
        return self._add_to_layout(opts.get('parent'), _.panes, pane)

    @may_handle(TmuxOpen)
    def open(self):
        name = self.msg.name
        opt = self.msg.options
        pinned = opt.get('pinned').exists(I)
        return (
            self._open_pane_ppm(name),
            (Nop() if pinned else TmuxPostOpen(name, opt - 'pinned'))
        )

    @handle(TmuxPostOpen)
    def open_pinned(self):
        def go(pane):
            return (
                self._pinned_panes(self.msg.ident) /
                L(TmuxOpen)(_, Map(pinned=True))
            ) + List(TmuxFixFocus(pane.ident), TmuxPack().pub.at(1))
        return self._pane(self.msg.ident) / go

    @handle(TmuxFixFocus)
    def fix_focus(self):
        return (
            self._pane(self.msg.pane) // (
                lambda a:
                a.focus.cata(Just(a), self._vim_pane) //
                self._focus_pane
            )
        )

    @handle(TmuxFocus)
    def focus(self):
        return self._pane(self.msg.pane) // self._focus_pane

    def _focus_pane(self, pane: Pane):
        return self.panes.find(pane) / _.focus / Task / UnitTask

    @may_handle(TmuxClosePane)
    def close(self):
        return self._close_pane_ppm(self.msg.name)

    @may_handle(TmuxRunCommand)
    def run_command(self):
        command = self.msg.command
        options = self.msg.options
        shell = options.get('shell').or_else(command.shell) // self.data.shell
        return shell.cata(
            L(self._run_in_shell)(command, _, options),
            L(self._run_command)(command, options)
        )

    @may_handle(TmuxRunShell)
    def run_shell(self):
        shell = self.msg.shell
        default = Map(shell.target.map(lambda a: ('target', a)).to_list)
        opts = default ** self.msg.options
        return TmuxRunCommand(command=shell, options=opts)

    @handle(SetShellTarget)
    def set_shell_target(self):
        target = self._pane(self.msg.target) / _.uuid
        setter = __.modify(__.set(target=target))
        return (self.data.command_lens(self.msg.shell.uuid)).map(setter)

    @handle(TmuxRunLineInShell)
    def run_line_in_shell(self):
        opt = self.msg.options
        return (
            opt.get('target') /
            (lambda a: Command(name='temp', line=a)) /
            L(self._run_in_shell)(_, self.msg.shell, opt)
        )

    @may_handle(ViewPathMod)
    def view_path_mod(self):
        return self._run_vpm(self.msg)

    @may_handle(PanePathMod)
    def pane_path_mod(self):
        return self._run_vpm(self.msg)

    @may_handle(LayoutPathMod)
    def layout_path_mod(self):
        return self._run_vpm(self.msg)

    @may_handle(Quit)
    def quit(self):
        t = self.state.possibly_open_panes // _.id / self.panes.close_id
        return QuitWatcher(), UnitTask(t.sequence(Task))

    @may_handle(QuitWatcher)
    def quit_watcher(self):
        return UnitTask(Task(self.machine.watcher.stop))

    @may_handle(TmuxInfo)
    def info(self):
        self.log.info('\n'.join(format_state(self.state)))

    @may_handle(Terminated)
    def terminated(self):
        return self._pane_mod(self.msg.pane.uuid, __.setter.pid(Empty()))

    @handle(SetCommandLog)
    def set_command_log(self):
        def set_log(cmd):
            log = lambda p: self.log.debug(
                'setting {} log to {}'.format(cmd.name, p))
            cmd_lens = self.data.command_lens(cmd.ident)
            log_path = self._pane(self.msg.pane_ident) // _.log_path
            log_path % log
            return cmd_lens / __.modify(__.set(log_path=log_path))
        return self._main_command(self.msg.cmd_ident) // set_log

    @may_handle(TmuxPack)
    def pack(self):
        return self._window_task(self.layouts.pack_window)

    @may_handle(TmuxMinimize)
    def minimize(self):
        return self._minimize(__.set(minimized=True))

    @may_handle(TmuxRestore)
    def restore(self):
        return self._minimize(__.set(minimized=False))

    @may_handle(TmuxToggle)
    def toggle(self):
        return self._minimize(lambda a: a.set(minimized=not a.minimized))

    @handle(TmuxOpenOrToggle)
    def open_or_toggle(self):
        name = self.msg.pane
        return (
            self._view(name) /
            self.views.is_open /
            __.cata(TmuxToggle(name), TmuxOpen(name))
        )

    def _minimize(self, f):
        return self._view_mod(self.msg.pane, f), TmuxPack().pub.at(1)

    def _window_task(self, f: Callable[[Window], Task[Either[str, Window]]]):
        return DataTask(_ // L(self._wrap_window)(_, f))

    def _wrap_window(self, data, callback):
        state = self._state(data)
        win = Task.from_maybe(state.vim_window, 'no vim window')
        update = __.map(L(self._with_vim_window)(data, state, _))
        return win // callback / update

    def _run_vpm(self, vpm):
        return self._window_task(vpm.run)

    def _open_pane_ppm(self, name: Ident):
        # cannot reference self.layouts.pack_path directly, because panes
        # are cached
        return _ident_vpm(name) / self._open_pane

    def _close_pane_ppm(self, name: Ident):
        return _ident_ppm(name) / self._close_pane / self._pack_path

    def _run_command_ppm(self, command, opt: Map):
        in_shell = Boolean('shell' in opt)
        pane_ident = opt.get('target') | self._default_pane_name
        kill = opt.get('kill') | command.kill
        signals = opt.get('signals') / List.wrap | command.signals
        def check_running(path):
            return Task.now(Right(path)) if in_shell else (
                self.panes.ensure_not_running(path.view, kill=kill,
                                              signals=signals) /
                __.replace(path)
            )
        def pipe(path):
            return (
                self.panes.pipe(path.view, self.state.instance_id) /
                path.setter.view /
                Right
            )
        def run(path):
            return (
                self._run_command_line(command.line, path.view, opt) /
                (lambda a: Right(path))
            )
        def pid(path):
            return (
                self.panes.command_pid(path.view) /
                (lambda a: path.map_view(__.set(pid=a))) /
                Right
            )
        def watch(path):
            msg = WatchCommand(command, path.view)
            return (
                Task.call(self.machine.watcher.send, msg)
                .replace(Right(path))
            )
        runner = (
            (self._open_pane_ppm(pane_ident) + check_running) /
            pipe /
            run /
            pid /
            watch
        )
        set_log = command.transient.no.maybe(
            SetCommandLog(command.uuid, pane_ident))
        return set_log.to_list.cons(runner).cat(TmuxPostOpen(pane_ident, opt))

    def _open_pane(self, path):
        return self.layouts.open_pane(path)

    def _close_pane(self, w):
        return self.layouts.close_pane_path(w)

    def _pack_path(self, path):
        return self.layouts.pack_path(path)

    def _layout_lens_bound(self, ident):
        ll = lambda l: Just(lens()) if l.has_ident(ident) else sub(l)
        sub = __.layouts.find_lens(ll).map(lens().layouts.add_lens)
        return (
            self.state.windows
            .find_lens(lambda a: ll(a.root) / lens().root.add_lens) /
            lens(self.state).windows.add_lens
        )

    def _add_to_layout(self, parent: Maybe[str], target: Callable, view: View):
        name = parent | 'root'
        return (
            self._layout_lens_bound(name) /
            target /
            __.modify(__.cat(view)) /
            self.with_sub
        )

    def _pane(self, ident: Ident):
        return self.state.pane(ident)

    def _view(self, ident: Ident):
        return self.state.find_view(ident)

    @property
    def _vim_pane(self):
        return self._pane('vim')

    def _pane_uuid(self, ident: Ident):
        return self._pane(ident) / _.uuid

    def _default_pane_name(self):
        return 'make'

    def _pane_mod(self, ident: Ident, f: Callable[[Pane], Pane]):
        return self._mod(_ident_ppm, ident, f)

    def _view_mod(self, ident: Ident, f: Callable[[View], View]):
        return self._mod(_ident_vpm, ident, f)

    def _mod(self, ctor: Callable, ident: Ident, f: Callable[[View], View]):
        def g(path):
            return Task.now(Right(path.map_view(f)))
        return ctor(ident) / g

    @property
    def _vim_pid(self):
        return self.vim.call('getpid').to_either('no pid')

    def _find_vim_pane(self, vim_pid):
        return self.server.pane_data.find(__.command_pid.contains(vim_pid))

    def _run_command_line(self, line, pane, options):
        return self.panes.run_command_line(pane, line)

    def _run_command(self, command, options):
        return self._run_command_ppm(command, options)

    def _run_in_shell(self, command: Command, shell: Shell, options: Map):
        ident = shell.target | self._default_pane_name
        opt = options + ('target', ident) + ('shell', shell)
        cmd_runner = self._run_command(command, opt)
        shell_runner = TmuxRunShell(shell, Map())
        return cmd_runner.cons(shell_runner)

    def _pinned_panes(self, ident):
        def layout(l):
            target = l.panes.find(__.has_ident(ident))
            sub = (l.layouts // layout)
            add = (List() if sub.empty and target.empty else
                   l.panes.filter(_.pin))
            return sub + add
        return (self.state.windows / _.root // layout) / _.ident

    def _main_command(self, cmd_ident: Ident):
        holder = lambda cmd: cmd.shell // self.data.shell | cmd
        return self.data.command(cmd_ident) / holder


class Plugin(MyoComponent):
    Transitions = TmuxTransitions

    @lazy
    def socket(self):
        return self.vim.vars.p('tmux_socket') | None

    @property
    def native_server(self):
        return NativeServer(socket_name=self.socket)

    @property
    def server(self):
        return Server(self.native_server)

    @property
    def layouts(self):
        return LayoutFacade(self.server)

    @property
    def panes(self):
        return PaneFacade(self.server)

    @property
    def views(self):
        return ViewFacade(self.server)

    def new_state(self):
        return TmuxState(server=self.server)

    @lazy
    def watcher(self):
        interval = self.vim.vars.p('tmux_watcher_interval') | 1.0
        return Watcher(self, interval=interval)

__all__ = ('Plugin',)



from ribosome import Machine, NvimFacade, RootMachine
from ribosome.nvim import HasNvim

from amino import List, _, L, Map
from ribosome.machine.state import SubMachine, SubTransitions

from myo.logging import Logging
from myo.env import Env
from myo.util.callback import parse_callback_spec
from myo.util.callback import VimCallback


class MyoHelpers:

    @property
    def options(self):
        return Map()

    def _callback_errors(self, spec, cb):
        def log(s, e):
            self.log.error('failed to parse callback \'{}\': {}'.format(s, e))
        (spec.zip(cb)).map2(lambda s, c: c.lmap(L(log)(s, _)))

    def _inst_callback(self, name, target):
        t = target or self.vim
        return name(t) if issubclass(name, VimCallback) else name

    def _inst_callbacks(self, spec, target):
        cb = spec / parse_callback_spec / _.join
        self._callback_errors(spec, cb)
        return cb.filter(_.is_right).join / L(self._inst_callback)(_, target)

    def _callback(self, name, target=None):
        spec = self.options.get(name).o(self.vim.vars.p(name))
        return self._inst_callbacks(spec, target)

    def _callbacks(self, name, target=None):
        t = target or self.vim
        def inst(name):
            return name(t) if issubclass(name, VimCallback) else name
        var = self.vim.vars.pl(name) | List()
        opt = self.options.get(name) | List()
        return self._inst_callbacks(var + opt, target)


class MyoComponent(SubMachine, MyoHelpers, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, parent=None, title=None) -> None:
        Machine.__init__(self, parent, title=title)
        HasNvim.__init__(self, vim)


class MyoState(RootMachine, Logging):
    _data_type = Env

    @property
    def title(self):
        return 'myo'


class MyoTransitions(SubTransitions, MyoHelpers, HasNvim, Logging):

    @property
    def options(self):
        return self.msg.options

    def __init__(self, machine, *a, **kw):
        SubTransitions.__init__(self, machine, *a, **kw)
        HasNvim.__init__(self, machine.vim)

__all__ = ('MyoComponent', 'MyoState', 'MyoTransitions')



from typing import Callable, Any

from amino.task import Task

from lenses import Lens, lens
from amino import Either, List, F, _, L, Left, __, Right, I
from amino.lens.tree import path_lens_unbound_pre

from ribosome.record import Record, field, list_field
from ribosome.machine import Message
from ribosome.machine.transition import Error

from myo.ui.tmux.window import Window
from myo.logging import Logging
from myo.ui.tmux.view import View
from myo.ui.tmux.layout import Layout


class ViewPath(Record):
    window = field(Window)
    view = field(View)
    layout = field(Layout)
    outer = list_field()

    @staticmethod
    def create(window, view, layout, outer):
        return ViewPath(window=window, view=view, layout=layout, outer=outer)

    @staticmethod
    def try_create(views: List[View]) -> Either[str, 'ViewPath']:
        def try_create2(view, win, layouts):
            return (
                layouts.detach_last.map2(F(ViewPath.create, win, view))
                .to_either('ViewPath.create: last item is not View')
            )
        def try_create1(view, layouts):
            return (
                layouts.detach_head
                .to_either('ViewPath.create: last item is not View')
                .flat_map2(F(try_create2, view))
            )
        return (
            views.detach_last.to_either('ViewPath.create: empty List[View]')
            .flat_map2(try_create1)
        )

    @property
    def layouts(self):
        return self.outer.cat(self.layout)

    @property
    def to_list(self):
        return self.layouts.cat(self.view).cons(self.window)

    def map(self, fp: Callable[[View], View], fl: Callable[[Layout], Layout]):
        return ViewPath.create(
            self.window, fp(self.view), fl(self.layout), self.outer / fl)

    def map_view(self, f: Callable[[View], View]):
        return self.map(f, I)

PPTrans = Callable[[ViewPath], Task[Either[Any, ViewPath]]]


class ViewPathLens(Record):
    lens = field(Lens)

    @staticmethod
    def create(lens: Lens):
        return ViewPathLens(lens=lens)

    def run(self, win: Window, f: PPTrans) -> Task[Either[Any, Window]]:
        bound = self.lens.bind(win)
        path = ViewPath.try_create(List.wrap(bound.get()))
        def update(pp):
            return bound.set(pp.to_list)
        def apply(pp):
            return f(pp) / (_ / update)
        return Task.from_either(path) // apply


def _initial_ppm_f(pp, window) -> Task[Either[Any, Window]]:
    return Task.now(Right(window))


class ViewPathMod(Logging, Message):
    pred = field(Callable)
    _f = field(Callable, initial=(lambda: _initial_ppm_f))

    @property
    def attr(self):
        return (lambda a: self._pane_attr(a).or_else(self._layout_attr(a)))

    @property
    def _pane_attr(self):
        return self._attr(_.panes)

    @property
    def _layout_attr(self):
        return self._attr(_.layouts)

    def _attr(self, f):
        return (
            f(__)
            .find_lens_pred(self.pred)
            .map(f(lens()).add_lens)
        )

    def _chain(self, f: PPTrans, g: Callable):
        h = lambda pp: L(pp.run)(_, f)
        chain = lambda pp, win: self._f(pp, win) // g(win, h(pp))
        return type(self)(pred=self.pred, _f=chain)  # type: ignore

    def map(self, f: PPTrans) -> Task[Either[Any, Window]]:
        keep_error = F(Left) >> Task.now
        return self._chain(f, lambda w, d: __.cata(keep_error, d))

    __truediv__ = map

    def and_then(self, f: PPTrans):
        return self._chain(f, lambda w, d: lambda a: d(a | w))

    __add__ = and_then

    def run(self, window: Window) -> Task:
        return self._lens(window) // L(self._f)(_, window)

    def _lens(self, window):
        sub = _.layouts
        pre = _.root
        return (
            Task.from_maybe(
                path_lens_unbound_pre(window, sub, self.attr, pre),
                Error('lens path failed for {}'.format(self.pred))
            ) /
            ViewPathLens.create
        )


class PanePathMod(ViewPathMod):

    @property
    def attr(self):
        return self._pane_attr


class LayoutPathMod(ViewPathMod):

    @property
    def attr(self):
        return self._layout_attr


def ppm_id(path: ViewPathMod):
    return Task.now(Right(path))

__all__ = ('ViewPath', 'ViewPathLens', 'ViewPathMod', 'ppm_id')



import abc

from amino import List, __
from amino.lazy import lazy

from myo.output.data import OutputLine, ParseResult
from myo.util.callback import VimCallback
from myo.state import MyoHelpers


class Reifier(VimCallback, MyoHelpers):

    @abc.abstractmethod
    def __call__(self, result: ParseResult) -> List[OutputLine]:
        ...

    @lazy
    def _truncator(self):
        return self._callback('path_truncator')

    def _truncate(self, path):
        return self._truncator / __(path) | path


class LiteralReifier(Reifier):

    def __call__(self, result):
        return result.lines

__all__ = ('Reifier', 'LiteralReifier')



from abc import ABCMeta, abstractmethod

from amino import __, Left, _, Map
from ribosome.machine import Message
from ribosome.record import list_field, Record

from myo.command import Command
from myo.logging import Logging


class Dispatcher(Logging, metaclass=ABCMeta):

    @abstractmethod
    def can_run(self, cmd: Command) -> bool:
        ...

    @abstractmethod
    def message(self, cmd: Command, options: Map) -> Message:
        ...


class Dispatchers(Record):
    dispatchers = list_field()

    def message(self, cmd: Command, options: Map):
        eligible = self.dispatchers.filter(__.can_run(cmd))
        if eligible.length > 1:
            msg = 'Multiple dispatchers for {}: {}'
            return Left(msg.format(cmd, eligible))
        else:
            return (
                eligible
                .head
                .to_either('Cannot dispatch {}'.format(cmd)) /
                __.message(cmd, options)
            )

    def __add__(self, dispatcher: Dispatcher):
        return self.append.dispatchers([dispatcher])

    def __getitem__(self, name):
        return self.dispatchers.find(_.name == name)

    def __str__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ','.join(map(repr, self.dispatchers))
        )

__all__ = ('Dispatchers', 'Dispatcher')



from myo.output.syntax.base import OutputSyntax
from myo.output.data import CodeEntry

from amino import Task, Map, _


class Syntax(OutputSyntax):

    @property
    def _handlers(self):
        return Map({
            CodeEntry: self._code,
        })

    def __call__(self, lines):
        return ((lines.with_index.flat_map2(self._line)).sequence(Task) +
                self._file)

    def _line(self, index, line):
        def run(entry):
            return (
                self._handlers.get(type(entry)) /
                (lambda a: a(index, line, entry))
            )
        return line.entry // run

    def _code(self, index, line, entry):
        col = (line.target.col / (_.col + 1) | 1) + line.indent
        rex = '\%{}c.'.format(col)
        return self._cont('ErrorMsg', self._line_re(index, rex), 'MyoCode')

    @property
    def _file(self):
        path = '.\+\ze '
        line = ' \zs\d\+'
        return (self._cont('MyoPath', path, 'MyoFile') +
                self._cont('MyoPath', line, 'MyoFile'))

__all__ = ('Syntax',)
